from argparse import ArgumentParser
from collections import defaultdict
from io import StringIO
from pathlib import Path
from numpy import nan
import pandas as pd
from shutil import rmtree
import typing as T
import yaml
import warnings

from parser import parse_mother_request


df = None

def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "path",
        help="Path for workflow. Serves as input and output.",
        type=str,
    )
    parser.add_argument(
        "--flat",
        action="store_true",
        help=(
            "Write application yaml as a single file, "
            "as opposed to in separate files (opa.yml, gsv.yml, meta.yml) "
            "in a same directory."
        ),
    )
    parser.add_argument(
        "--no-abstract",
        action="store_true",
        help=(
            "Do not abstract variables, i.e. have a directory per application variable, "
            "as opposed to having a single `gsv.yml` and `opa.yml` file with abstracted (common) "
            "variables as header"
        ),
    )
    parser.add_argument(
        "--clear",
        action="store_true",
        help="Remove content of created directories if they already existed."
    )
    return parser


def _row_to_dict(row: pd.Series, prefix: str = ""):
    # Pandas row to dict
    dct = row.to_dict()
    # Remove unused fields
    dct = {key: value for key, value in dct.items() if value is not nan}
    # Remove prefix from keys
    if prefix:
        dct = {key.removeprefix(prefix): value for key, value in dct.items()}
    return dct


def create_yaml_dicts(path: Path):
    global df  # pylint: disable=W0603

    mother_request = path / "conf" / "mother_request.yml"
    df = parse_mother_request(mother_request)
    opa_columns = [col for col in df.columns if col.startswith("opa_")]
    gsv_columns = [col for col in df.columns if col.startswith("gsv_")]
    dcts = defaultdict(dict)
    for app, app_df in df.groupby("Application"):
        for i, (_, request) in enumerate(app_df.iterrows(), start=1):
            dcts[app][i] = {
                    "GSVREQUEST": _row_to_dict(request[gsv_columns], "gsv_"),
                    "OPAREQUEST": _row_to_dict(request[opa_columns], "opa_"),
            }
    return dcts


def write_yamls_flat(path: Path):
    yamls = create_yaml_dicts(path)
    yamls_path = path / "conf" / "applications"
    for app, yaml_content in yamls.items():
        yaml_filename = f"{app.lower()}.yml"
        yaml_filepath = yamls_path / yaml_filename
        yaml_io = StringIO()
        yaml.safe_dump(yaml_content, yaml_io, default_flow_style=False)
        with open(yaml_filepath, "w", encoding="utf-8") as wf:
            wf.write(
                "---\n\n"
            )
            wf.write(
                yaml_io.getvalue().replace("\'", "\"")
            )


def write_yamls_tree(path: Path, clear: bool = False):
    yamls = create_yaml_dicts(path)
    conf_apps_path = path / "conf" / "applications"
    request_filename = [
        ("GSVREQUEST", "gsv.yml"),
        ("OPAREQUEST", "opa.yml"),
    ]
    for app, app_requests in yamls.items():
        app_path = conf_apps_path / app.lower()
        if request_path.exists() and clear:
            rmtree(app_path)
        app_path.mkdir(exist_ok=True, parents=False)
        print(f"Made dir: {app_path}")
        for i, request in app_requests.items():
            request_path = app_path / f"{i:02d}"
            request_path.mkdir(exist_ok=True, parents=False)
            for req, filename in request_filename:
                with open(request_path / filename, "w", encoding="utf-8") as wf:
                    print(request[req])
                    yaml.dump(request[req], wf)
        with open(app_path / "meta.yml", "w", encoding="utf-8") as wf:
            yaml.dump({
                "APPNAME": app,
            })


def _len_unique(values: T.Sequence[T.Any]):
    if any(isinstance(element, list) for element in values):
        values = [
            (element if not isinstance(element, list) else tuple(element))
            for element in values
        ]
    return len(set(values))


def _sort_by_count(lst):
    def sort(x):
        return lst.count(x)
    return sorted(lst, key=sort)


def write_yamls_abstract(path: Path, clear: bool = False):
    def update(dct_lists: T.Dict[str, T.List[T.Any]], dct: T.Dict[str, T.Any]):
        for key, value in dct.items():
            dct_lists[key].append(value)

    def split_unique_non_unique(dct_lists):
        num_requests = len(list(dct_lists.values())[0])
        unique = {
            key: values[0] for key, values in dct_lists.items() if _len_unique(values) == 1
        }
        non_unique = []
        for i in range(num_requests):
            current_request = {}
            for key, values in dct_lists.items():
                if key in unique:
                    continue
                try:
                    current_request.update({key: values[i]})
                except IndexError:
                    top_value = _sort_by_count(values)[0]
                    warnings.warn(
                        f"Application {app} makes up {i}-th request: <{key}: {top_value}>."
                    )
                    current_request.update({key: top_value})
            non_unique.append(current_request)
        return unique, non_unique

    def dump(dct_lists, path):  # pylint: disable=W0621
        content, non_unique_keys = split_unique_non_unique(dct_lists)
        if non_unique_keys:
            content.update({
                "REQUESTS": non_unique_keys,
            })
        with open(path, "w", encoding="utf-8") as wf:
            yaml.dump(content, wf)

    yamls = create_yaml_dicts(path)
    conf_apps_path = path / "conf" / "applications"

    for app, app_requests in yamls.items():
        app_path = conf_apps_path / app.lower()
        if app_path.exists() and clear:
            rmtree(app_path)
        app_path.mkdir(exist_ok=True, parents=False)

        # Keys are opa request keys, values are lists with one entry per request
        opa_requests = defaultdict(list)
        gsv_requests = defaultdict(list)

        for _, request in app_requests.items():
            update(gsv_requests, request["GSVREQUEST"])
            update(opa_requests, request["OPAREQUEST"])

        dump(opa_requests, app_path / "opa.yml")
        dump(gsv_requests, app_path / "gsv.yml")

        with open(app_path / "meta.yml", "w", encoding="utf-8") as wf:
            yaml.dump({
                "APPNAME": app,
            }, wf)


if __name__ == "__main__":
    args = _get_parser().parse_args()

    if args.flat:
        write_yamls_flat(Path(args.path))
    else:
        if args.no_abstract:
            write_yamls_tree(Path(args.path))
        else:
            write_yamls_abstract(Path(args.path), args.clear)
